local global = require "eatglobal_v1_3"
local utils = global.utils
local cargocalculatedweight_config = require "eat1963_cargocalculatedweight_config"

local defaultUserSettings = {
	disableTpfDivFour = {
		type ="boolean",
		name = _("name_eat1963_CCW_disableTpfDivFour"),
		description = _("desc_eat1963_CCW_disableTpfDivFour"),
		default = false,
	},
	calcUnknownVehicle = {
		type ="boolean",
		name = _("name_eat1963_CCW_calcUnknownVehicle"),
		description = _("desc_eat1963_CCW_calcUnknownVehicle"),
		default = true,
	},
	useRealWeights = {
		type ="boolean",
		name = _("name_eat1963_CCW_useRealWeights"),
		description = _("desc_eat1963_CCW_useRealWeights"),
		default = true,
	},
	tryCalcLoadSpeedAutomatically = {
		type ="boolean",
		name = _("name_eat1963_CCW_tryCalcLoadSpeedAutomatically"),
		description = _("desc_eat1963_CCW_tryCalcLoadSpeedAutomatically"),
		default = false,
	},
	cityRequirementsFactor = {
		type ="number",
		name = _("name_eat1963_CCW_cityRequirementsFactor"),
		description = _("desc_eat1963_CCW_cityRequirementsFactor"),
		default = 1,
		min = 0.5,
		max = 8,
	},
	capacityFactorGoods = {
		type ="number",
		name = _("name_eat1963_CCW_capacityFactorGoods"),
		description = _("desc_eat1963_CCW_capacityFactorGoods"),
		default = 1,
		min = 0.25,
		max = 10,
	},
	capacityFactorPassenger = {
		type ="number",
		name = _("name_eat1963_CCW_capacityFactorPassenger"),
		description = _("desc_eat1963_CCW_capacityFactorPassenger"),
		default = 1,
		min = 0.25,
		max = 10,
	},
	loadSpeedFactor = {
		type ="number",
		name = _("name_eat1963_CCW_loadSpeedFactor"),
		description = _("desc_eat1963_CCW_loadSpeedFactor"),
		default = 1,
		min = 0.25,
		max = 10,
	},
}

--- Erzeugt bzw. öffnet eine globale Konfiguration
-- gibt Tabelle mit Konfiguration zurück
local function getConfig ()
	local id = "cargo_calculated_weight"
	if global.configs.existsConfig (id) then
		return global.configs.getConfig (id)
	else
		local result = global.configs.createConfig (id, cargocalculatedweight_config)
		result.initializeUserConfig(defaultUserSettings)
		return result
	end
end

local function getTransportVehicle ( data )
	if data and data.metadata and data.metadata.transportVehicle then
		return data.metadata.transportVehicle
	else
		return nil
	end
end

--- Prüft, ob Modell eine bestimmte Güterart transportieren kann
-- @param data, die Daten vom Modell
-- @param id, die Art von Gütern
-- @return true, wenn Modell Güterart transportieren kann
local function hasGood ( data , id )
	if data and data.metadata and data.metadata.transportVehicle then
		if data.metadata.transportVehicle.capacities then
			for i = 1, #data.metadata.transportVehicle.capacities do
				if data.metadata.transportVehicle.capacities[i].type == id then
					return true
				end
			end
		elseif data.metadata.transportVehicle.compartments then
			for i = 1, #data.metadata.transportVehicle.compartments do
				for j = 1, #data.metadata.transportVehicle.compartments[i] do
					for k = 1, #data.metadata.transportVehicle.compartments[i][j] do
						if data.metadata.transportVehicle.compartments[i][j][k].type == id then
							return true
						end
					end
				end
			end
		end
	end

	return false
end

--- Ermittelt die maximal mögliche Kapazität eines Fahrzeuges
local function getMaxCapacity ( data )
	local calcData = getConfig ().cargodata.calcData
	local capacity = 0
	local transportVehicle = getTransportVehicle (data)
	
	if transportVehicle then
		if transportVehicle.capacities then
			for i = 1, #transportVehicle.capacities do
				capacity = math.max(capacity, transportVehicle.capacities[i].capacity or calcData.defaultCapacity)
			end
		elseif transportVehicle.compartments then
			for i = 1, #data.metadata.transportVehicle.compartments do
				for j = 1, #data.metadata.transportVehicle.compartments[i] do
					for k = 1, #data.metadata.transportVehicle.compartments[i][j] do
						capacity = math.max(capacity, transportVehicle.compartments[i][j][k].capacity or calcData.defaultCapacity)
					end
				end
			end
		end
	end
	
	return capacity
end

--- Berechnet den loadSpeed
-- @param data, data des Fahrzeuges
-- @return, berechneten loadSpeed anhand der Benutzereinstellungen
local function getLoadSpeed ( data )
	local config = getConfig ()
	local result = 2
	
	if config.userConfig.tryCalcLoadSpeedAutomatically then
		local capacity = getMaxCapacity (data)
		local transportVehicle = getTransportVehicle (data)
		if transportVehicle and transportVehicle.compartments then
			capacity = capacity * #transportVehicle.compartments
		end
		if config.userConfig.disableTpfDivFour then
			result = capacity / (config.cargodata.calcData.calcLoadSpeedFaktor * config.cargodata.calcData.tpfDivFourFactor)
		else
			result = capacity / config.cargodata.calcData.calcLoadSpeedFaktor
		end
	else
		if data and data.transportVehicle then
			result = data.transportVehicle.loadSpeed or 2
		end
	end
	result = result * config.userConfig.loadSpeedFactor
	
	return utils.round (result)
end

---
local function calcCapacitys (transportVehicle, maxVolume, maxWeight)
	local config = getConfig ()
	local calcData = config.cargodata.calcData
	local userConfig = config.userConfig
	local capacity = 0
	local singleWeight = 0
	
	-- Hilfsfunktion: Gewicht des Gütertyps in Tonnen ermitteln
	local function getCargoWeight ( cargoType )
		result = config.cargodata.weights[cargoType] or config.cargodata.calcData.defaultTpfWeight
		return result / 1000
	end
	
	if transportVehicle.capacities then
		for i = 1, #transportVehicle.capacities do
			if not config.cargodata.locked.cargoTypes[transportVehicle.capacities[i].type] then
				singleWeight = getCargoWeight( transportVehicle.capacities[i].type )
				capacity = math.min (maxWeight / singleWeight, maxVolume)
				if config.userConfig.disableTpfDivFour then
					transportVehicle.capacities[i].capacity = capacity * calcData.tpfDivFourFactor * userConfig.capacityFactorGoods
				else
					transportVehicle.capacities[i].capacity = capacity * userConfig.capacityFactorGoods
				end
			end
		end
	elseif transportVehicle.compartments then
		local countCompartments = #transportVehicle.compartments
		for i = 1, #transportVehicle.compartments do
			for j = 1, #transportVehicle.compartments[i] do
				for k = 1, #transportVehicle.compartments[i][j] do
					if not config.cargodata.locked.cargoTypes[transportVehicle.compartments[i][j][k].type] then
						singleWeight = getCargoWeight( transportVehicle.compartments[i][j][k].type )
						capacity = math.min (maxWeight / singleWeight, maxVolume)
						-- wir müssen die Anzahl der compartments berücksichtigen
						capacity = math.max(utils.round(capacity / countCompartments), 1)
						if config.userConfig.disableTpfDivFour then
							transportVehicle.compartments[i][j][k].capacity = capacity * calcData.tpfDivFourFactor * userConfig.capacityFactorGoods
						else
							transportVehicle.compartments[i][j][k].capacity = capacity * userConfig.capacityFactorGoods
						end
					end
				end
			end
		end
	end
end

--- Führt alle nötigen Berechnungen für Passagierfahrzeuge durch
-- Passagierfahrzeuge werden etwas anders behandelt, daher diese Function
local function handlePassengers ( data, fileName )
	local config = getConfig ()
	
	if data and data.metadata and data.metadata.transportVehicle and data.metadata.transportVehicle.capacities and
		(not config.isLocked (config, fileName, data.metadata.transportVehicle)) then
	 
		local capacity = data.metadata.transportVehicle.capacities[1].capacity
		
		-- Kapazität
		if config.userConfig.disableTpfDivFour then
			capacity = capacity * config.cargodata.calcData.tpfDivFourFactor
		end
		capacity = capacity * config.userConfig.capacityFactorPassenger
		
		data.metadata.transportVehicle.capacities[1].capacity = capacity
		data.metadata.transportVehicle.loadSpeed = getLoadSpeed (data)
	end

	return data
end

--- Behandlung von Güterfahrzeugen
--
local function handleGoods ( data, fileName )
	local config = getConfig ()

	if data and data.metadata and data.metadata.transportVehicle and
		(data.metadata.transportVehicle.capacities or data.metadata.transportVehicle.compartments) and
			(not config.isLocked (config, fileName, data.metadata.transportVehicle)) then

		if (not data.metadata.transportVehicle.calculated) or (config.userConfig.calcUnknownVehicle) then
			--- max. Ladevolumen des Fahrzeuges ermitteln
			-- es steht entweder unter data.metadata.transportVehicle.maxVolume oder, falls nicht,
			-- wird die (momentan) maximal mögliche Kapazität als Bezug genommen
			local maxVolume = data.metadata.transportVehicle.maxVolume or getMaxCapacity (data)
			--- max. Nutzlast ermitteln
			-- zu finden unter data.metadata.transportVehicle.maxWeight oder aus maxVolume errechnen
			local maxWeight = data.metadata.transportVehicle.maxWeight or maxVolume / (config.cargodata.calcData.defaultTpfWeight / 1000)
			calcCapacitys (data.metadata.transportVehicle, maxVolume, maxWeight)
		end
		data.metadata.transportVehicle.loadSpeed = getLoadSpeed (data)
	end

	--config.logFn ("(handleGoods) ["..fileName.."] data.metadata.transportVehicle", data.metadata.transportVehicle)

	return data
end

local function initLockedCargoTypes ()
	local config = getConfig ()
	
	local function __isLocked( cargoType )
		for i = 1, #config.cargodata.locked.cargoTypes do
			if config.cargodata.locked.cargoTypes[i] == cargoType then
				return true
			end
		end
		return false
	end
	
	local result = { }
	for i = 1, #game.config.cargotypes do
		result[game.config.cargotypes[i].id] = __isLocked( game.config.cargotypes[i].id )
	end
	config.cargodata.locked.cargoTypes = result
end

local function initCargoWeights ()
	local config = getConfig ()
	if not config.cargodata.initialized then
		for i = 1, #game.config.cargotypes do
			if (game.config.cargotypes[i].id == "PASSENGERS") then
				Compactness = (config.cargodata.cargoCompactness[game.config.cargotypes[i].id] or 0) / config.userConfig.capacityFactorPassenger
			else
				Compactness = (config.cargodata.cargoCompactness[game.config.cargotypes[i].id] or 0) / config.userConfig.capacityFactorGoods
			end
			if Compactness > 0 then
					if config.userConfig.useRealWeights then
						game.config.cargotypes[i].weight = Compactness
					else
						game.config.cargotypes[i].weight = Compactness * config.cargodata.calcData.cargoFactor
					end
				-- Liste mit den bereits kalkulierten Werten für schnellen Zugriff bei evtl. dynamischen Berechnungen 
				config.cargodata.weights[game.config.cargotypes[i].id] = game.config.cargotypes[i].weight
			end
		end
		config.cargodata.initialized = true
		config.logFn ("config.cargodata.weights", config.cargodata.weights)
	end
end

local function fillVehicleData (fileName, data)
	local lowerFileName = string.lower(fileName)
	local config = getConfig ()
	local result = { }

	local transportVehicle = getTransportVehicle (data)
	if transportVehicle and (transportVehicle.capacities or transportVehicle.compartments) then
	--
		local daten = config.calcWeightAndVolume (config, fileName, transportVehicle)
		if daten then
			transportVehicle.maxWeight = daten.maxWeight
			transportVehicle.maxVolume = daten.maxVolume
			transportVehicle.calculated = daten.calculated or false
			transportVehicle.isLocked = daten.isLocked or false
		end
		
		--config.logFn ("(fillVehicleData) ["..fileName.."] transportVehicle", transportVehicle)
		config.logFn ("(fillVehicleData) ["..fileName.."] daten", daten)
	end
	return data
end

local function applyUserPrefences ()
	local config = getConfig ()
	
	local function importLockedModels ( data )
		for i = 1, #data do
			if type(data[i]) == "string" then
				config.cargodata.locked.models[#config.cargodata.locked.models + 1] = data[i]
			end
		end
	end
	
	if not config.modSettings.userSettingsIsTransfered then
		-- Benutzereisntellungen aus user_config.lua einlesen
		local userConfig = utils.readData (config.env.scriptsPath.."user_config.lua")
		
		if not utils.fileExists (config.env.modPath.."settings.lua") then
			-- Ok, der User hat noch keine Einstellungen in settings.lua vorgenommen (im TPFMM?!)
			-- und wir müssen die alten Einstellungen transferieren
			
			-- Extrabehandlung für geänderte(n) Wert(e)
			if (userConfig.useOnlyKnownData ~= nil) then
				userConfig.calcUnknownVehicle = not userConfig.useOnlyKnownData
			end
			if (userConfig.capacityFactor ~= nil) then
				userConfig.capacityFactorGoods = userConfig.capacityFactor
			end
			-- Ohne Usereinstellungen in settngs.lua werden diese aus user_config.lua importiert und
			-- anschliessend in settings.lua abgespeichert
			config.importUserConfig(userConfig)
			config.saveUserConfig()
		end
		
		if not utils.fileExists (config.env.modPath.."blocked_models.lua") then
			-- Es existiert keine blocked_models.lua. Also import der blockierten Modelle aus Userconfig.
			importLockedModels (userConfig.lockedModels or {})
			-- Abspeichern in blocked_models.lua
			utils.saveData(userConfig.lockedModels, config.env.modPath.."blocked_models.lua")
		end
		-- user_config.lua wird nicht mehr benötigt, also löschen
		utils.deleteFile(config.env.scriptsPath.."user_config.lua")
		-- Flag setzen und speichern, dass Konvertierung erfolgt ist
		config.modSettings.userSettingsIsTransfered = true
		config.saveModSettings ()
	else
		-- Daten aus blocked_modells.lua impoortieren (gesperrte Modelle)
		local blocked_models = {}
		if utils.fileExists (config.env.modPath.."blocked_models.lua") then
			blocked_models = utils.readData (config.env.modPath.."blocked_models.lua")
		else
			-- Wenn keine blocked_models.lua existiert, legen wir eine leere Datei an.
			-- So kann der User diese finden, falls er sie editieren möchte.
			utils.saveData(blocked_models, config.env.modPath.."blocked_models.lua")
		end
		importLockedModels (blocked_models)
	end
	initLockedCargoTypes ()
	config.logFn ("(applyUserPrefences) modSettings", config.modSettings)
	config.logFn ("(applyUserPrefences) userSettings", config.userConfig)
	config.logFn ("(applyUserPrefences) config.cargodata.locked.models", config.cargodata.locked.models)
end

local function constructionHandler (fileName, data)
	local config = getConfig ()
	local fn = data.updateFn
	
	data.updateFn = function(params)
		result = fn(params)
		if data.type == "TOWN_BUILDING" and (data.townBuildingParams.landUseType == "INDUSTRIAL" or data.townBuildingParams.landUseType == "COMMERCIAL") then
			if result.stocks and #result.stocks > 1 then
				local capacity = utils.round (result.stocks[1].moreCapacity / config.userConfig.cityRequirementsFactor)
				if capacity > 0 then
					for i = 1, #result.stocks do
						result.stocks[i].moreCapacity = capacity
						result.stockRules[i].capacity = capacity
					end
				else
					table.remove(result, stocks)
					table.remove(result, stockRules)
				end
			end
		end
		return result
	end
	
	return data
end

local function modelHandler(fileName, data)
	if hasGood (data, "PASSENGERS") then
		data = handlePassengers (data, fileName)
	else
		-- Füllt Fahrzeudaten auf (Volumen, Nutzlast)
		data = fillVehicleData (fileName, data)
		-- Kapazitäten berechnen und zuweisen
		data = handleGoods (data, fileName)
	end
	
	return data
end

local function initModifier ()
	local config = getConfig ()
	-- Modifier (für die Modelle) installieren
	global.modifier.addModifier ( "all_cargo_at_all_vehicle", config.getId (),
		config.modifierSequenceList, "loadModel", modelHandler)

	if config.userConfig.cityRequirementsFactor ~= 1 then
		addModifier( "loadConstruction", constructionHandler )
	end
	
end

---- Führt Grund-Initialisierung durch
-- Wird/Muss in der runFn aufgerufen werden, da dann sichergestellt ist, 
-- dass das Spiel hinreichend initialisiert wurde
local function init ()
	-- Benutzereinstellungen einlesen
	applyUserPrefences ()
	-- Globale Gewichte der Güter berechnen
	initCargoWeights ()
	-- 
	initModifier ()
	
	--local config = getConfig ()
	--config.logFn ("config", config)
end

function data()
return {
  info = {
    minorVersion = 8,
    severityAdd = "NONE",
    severityRemove = "NONE",
    name = _("Cargo Calculated Weight"),
    description = _("Berechnet das Gewicht der Güterarten anhand der spezifischen Dichte."),
    authors = {
			{
          name = "EAT1963",
          role = "CREATOR",
          text = "Idee, Umsetzung",
          steamProfile = "76561198052411779",
          tfnetId = 19725,
      },
			{
          name = "PMV",
          role = "CREATOR",
          text = "Idee, Umsetzung",
          --steamProfile = "76561198052411779",
          tfnetId = 20103,
      },
    },
    tags = { "script" },
		tfnetId = 2872,
		url = "https://www.transportfever.net/filebase/index.php/Entry/2872-Cargo-Calculated-Weight/"
  },
	settings = defaultUserSettings,
	options = {},
	
	--[[=============================================]]
	
	runFn = function (settings)
		-- Grundinitialisierung(en) durchführen
		init ()
		-- Gewichtssimulation aktivieren
		game.config.simulateCargoWeight = true
		-- Gewichtssimulation auch für bestehende Savegames aktivieren
		game.config.simulateCargoWeightLegacy = true
	end
}
end
