--- Utils
-- @author Enno Sylvester
-- @copyright 2017
-- @module eatglobal.utils

local utils = {
	version = 1.3,
}

--[[
			##########################################
			environment-tools
			##########################################
]]

local function init()
	if (type(utils.env) ~= "table") then
		utils.env = {}
	end
	utils.env.isWinOS = (os.getenv("OS") == "Windows_NT")
end

init ()

function utils.getModulFileName ( level )
	if (level == nil) then
		level = 1
	end
	local info = debug.getinfo(level, "S")
	if (info ~= nil) then
		return string.gsub(string.gsub(info.source, '\\', '/'), '@', "")
	else
		return ""
	end
end

function utils.getModPath ()
	local function debugPath ()
		local level = 1
		local info
		
		repeat
			local info = debug.getinfo(level, "S")
			level = level + 1
		until (info == nil)
		return string.gsub(debug.getinfo(level - 2,'S').source, '\\', '/')
	end

	local result = debugPath ()
	if string.ends(result, 'mod.lua') then
    result = string.gsub(result, "@(.*/)mod[.]lua", "%1")
  else
    result = string.gsub(result, "@(.*/)res/*.+[.]*", "%1")
  end

	return result
end

--[[
			##########################################
			table-tools
			##########################################
]]

--- Kopiert eine Tabelle incl. der metatable
-- @param tbl
-- @return Kopie von tbl
-- Achtung: In Kopien von Proxytabellen kann nicht geschrieben oder gelöscht werden.
function utils.copyTable (tbl)
	local function doCopy (t)
		local result
		if (type(t) == 'table') then
			result = {}
			for key, value in next, t, nil do
				result[doCopy(key)] = doCopy(value)
			end
			setmetatable(result, doCopy(getmetatable(t)))
		else -- number, string, boolean, etc
			result = t
		end
		return result
	end
	return doCopy(tbl)
end

--- Create a ordered table
-- @return {table} a table with ordered key value iteration
-- @author Oskar Eisemuth
-- @copyright 2017
 function utils.createOrderedTable()
	local t = {}
	local mt = {
		__index = {
			_keyorder = {},
			_delete = function(self, key)
				if self[key] == nil then
					return
				end
				self[key] = nil
				local keyorder = self._keyorder
				for i, k in ipairs(keyorder) do
					if k == key then
						table.remove(keyorder, i)
					end
				end
			end
		},
		__newindex = function(self, key, value)
			rawset(self, key, value)
			local keyorder = self._keyorder
			keyorder[#keyorder+1] = key
		end,
		__pairs = function(tbl)
			local i = 0
			local function pairs_iter(self)
				i = i + 1
				local key = self._keyorder[i]
				if (key) then
					return key, self[key]
				end
				return
			end
			return pairs_iter, tbl, nil
		end
	}
	setmetatable(t, mt)
	return t
end

--- Erzeugt eine Proxy-Tabelle.
--	Alle Zugriffe werden direkt auf den 'master' umgeleitet
-- @utils.proxyTable
-- @param master
-- @return proxy
function utils.proxyTable(master)
  local metatable = {
    __index = master,
    __newindex = master,
		
		__len = function (tbl)
			return #master
		end,
		
		__pairs = function (tbl)
      local function iterFn(tbl, k)
        local v
        -- Implement your own key,value selection logic in place of next
        k, v = next(master, k)
        if v then
          return k,v
        end
      end    
      -- Return an iterator function, the table, starting point
      return iterFn, tbl, nil
    end,
    
		__ipairs = function (tbl)
			-- Iterator function
			local function iterFn(tbl, i)
				-- Implement your own index, value selection logic
				i = i + 1
				local v = master[i]
				if v then
					return i, v
				end
			end
			-- return iterator function, table, and starting point
			return iterFn, tbl, 0
    end,
  }
  local proxy = {}
  setmetatable(proxy, metatable)
  return proxy
end

--- Äquivalent zu pairs
-- Iteriert über eine Tabelle, wobei die Keys in sortierter Reihenfolge aufgerufen werden.
-- Nutzung:
-- for key, value in sortPairs (t) do
-- end
function utils.sortPairs ( t, order )
	-- collect the keys
	local keys = {}
	for k in pairs(t) do
		keys[#keys + 1] = k
	end

	-- if order function given, sort by it by passing the table and keys a, b,
	-- otherwise just sort the keys 
	if order then
		table.sort(keys, function(a,b) return order(t, a, b) end)
	else
		table.sort(keys)
	end

	-- return the iterator function
	local i = 0
	return function()
		i = i + 1
		if keys[i] then
			return keys[i], t[keys[i]]
		end
	end
end

--[[
			##########################################
			log- and filetools
			##########################################
]]

--- log mit Präfix
function utils.createLogFn (prefix)
	local prefix = prefix..": "
	local logfn = function (...)
		local params = {prefix, ...}
		local stack = {}
		for i = 1, #params do
			stack[#stack + 1] = utils.toString(params[i])
		end
		print(table.concat(stack))
	end
	return logfn
end

--- prüft, ob eine Datei vorhanden ist
-- @utils.fileExists
-- @param fileName
-- @return true, wenn vorhanden
function utils.fileExists (fileName)
	local file = io.open(fileName, "r")
	if file then
		file:close()
		return true
	end
	return false
end

--- Daten aus Datei einlesen
-- @utils.readData
-- @param fileName
-- @return Tabelle mit eingelesene Datan oder eine leere Tabelle
function utils.readData (fileName)
	local file, err = loadfile(fileName)
	if (file ~= nil) then
		return file ()
	else
		return {}, err
	end
end

--- Daten in Datei schreiben
-- @utils.saveData
-- @param data
-- @param fileName
-- Format: siehe Comment serialize.serializeReturnfmt
function utils.saveData (data, fileName)
	local file = io.open(fileName, "w")
	if file then
		local s = utils.eatglobal.serialize.serializeReturnfmt(data)
		file:write(s)
		file:close()
	end
end

function utils.deleteFile (fileName)
	local function quote( var )
		return '"'..string.gsub(var, '/', '\\')..'"'
	end
		
	if utils.fileExists (fileName) then
		local cmd = nil
		if utils.env.isWinOS then
			cmd = '"del '..quote(fileName)..'" '
		else
			cmd = '"rm '..quote("./"..fileName)..'" '
		end
		p = io.popen(cmd)
		if p then
			p:close()
		end
	end
end

--[[
			##########################################
			math-tools
			##########################################
]]

function utils.getIndex(data, params, compareFn)
	for i = 1, #data do
		if compareFn (data[i], params) then
			return i
		end
	end
	return 0
end

function utils.getArrayIndex (array, params, compareFn)
	for index, value in ipairs(array) do
		if compareFn (value, params) then
			return index
		end
	end
	return 0
end

function utils.toString (var)
	if type(var) == "nil" then
		return "nil"
	elseif type(var) == "boolean" then
		return tostring(var)
	elseif type(var) == "number" then
		return tostring(var)
	elseif type(var) == "string" then
		return var
	elseif type(var) == "table" then
		return utils.eatglobal.serialize.serializeBlanc(var)
	else
		return ""
	end
end

function utils.round (var)
	return math.floor (var + 0.5)
end

--[[
			##########################################
			misc-tools
			##########################################
]]

return utils